<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN" "http://www.w3.org/TR/html4/loose.dtd">
<html>
	<head>
		<meta http-equiv="Content-Type" content="text/html; charset=UTF-8">
		<title>[% page.title %]</title>
		<meta name='Generator' content='Zim [% zim.version %]'>
		<style type='text/css'>
			a          { text-decoration: none      }
			a:hover    { text-decoration: underline }
			a:active   { text-decoration: underline }
			strike     { color: grey                }
			u          { text-decoration: none;
			             background-color: yellow   }
			tt         { color: #2e3436;            }
			pre        { color: #2e3436;
			             margin-left: 20px          }
			h1         { text-align: center;
			             color: #4e9a06             }
			h2         { color: #4e9a06             }
			h3         { color: #4e9a06             }
			h4         { color: #4e9a06             }
			h5         { color: #4e9a06             }
			span.insen { color: grey                }
			.page { max-width: 1000px;}
			.menu{
				float:left; width: 300px;
			}
			
			.content { padding-left: 320px;}
			.notebook{font-variant: small-caps;
					color:#4e9a06;
					padding: 0px 20px;}
			hr{clear:both;}
		</style>
		
	</head>
	<body>
		<div class="page">
			<div class="heading">
				<h1><a href="[% url(pages.home.name) %]" class="notebook">[% notebook.name %]:</a>
			
				
				[% IF page.properties.type == 'namespace-index' -%]
					Document Index</h1>
				[%- ELSE -%]
					[% page.heading %]</h1>
				[%- END %]
			</div>
			<hr>
			<div class="menu">
				[% pageindex() %]
			</div>
			<div class="content">
			<!-- Wiki content -->
				[% page.body %]

			<!-- End wiki content -->
			</div>
			<hr>
			<!-- Backlinks -->
			<div class="footer">
				[% IF page.backlinks -%] Backlinks:
					[%- FOREACH link = page.backlinks -%]
						<a href='[% url(link) %]'>[% link.name %]</a></li>
					[%- END -%]
				[%- ELSE -%]
					No backlinks to this page.
				[%- END %]
			<!-- End Backlinks -->
			</div>
		</div>
	</body>
</html>
