<html>
	<head>
		<meta http-equiv="Content-Type" content="text/html; charset=UTF-8">
		<title>[% page.title %]</title>
		<meta name='Generator' content='Zim [% zim.version %]'>
		<style type='text/css'>
			body     { background-color: #fff;
			           color: #000;
			           font-size: large              }
			a        { text-decoration: none         }
			a:hover  { text-decoration: underline    }
			a:active { text-decoration: underline    }
			strike   { color: grey                   }
			u        { text-decoration: none;
			           background-color: yellow      }
			pre      { background-color: lightyellow }
			.bar     { background-color: #E0E0FF;
			           margin: 0                     }
			.nav     { color: grey;
			           font-size: small              }
		</style>
		<script type='text/javascript'>
		<!--
// Slideshow script - adapted from old kwiki package

function nextSlide() {
	window.location="[% url(pages.next) %]"
}

function prevSlide() {
	window.location="[% url(pages.previous) %]"
}


function handleKey(e) {
	var key;
	if (e == null) { // IE
		key = event.keyCode
	}
	else { // Mozilla
		if (e.altKey || e.ctrlKey) {
			return true
		}
		key = e.which
	}

	switch(key) {
		case 8: prevSlide(); break // Backspace
		case 13: nextSlide(); break // Enter
		case 32: nextSlide(); break // Space
		default:
	}

	return false
}

document.onkeypress = handleKey
// document.onclick = nextSlide
// document.ondblclick = prevSlide

// End slideshow script
		// -->
		</script>
	</head>
	<body>

	<table width="100%" height="100%" border="0" cellpadding="5" cellspacing="0">
		<tr>
			<td colspan="3" align="center" valign="middle" class="bar">
				[% IF page.properties.type == 'namespace-index' -%]
				<h3 class='bar'>Table Of Contents</h3>
				[% ELSE %]
				<h3 class='bar'>[% page.title %]</h3>
				[% END %]
			</td>
		</tr>
		<tr>
			<td width="5%">&nbsp;</td>
			<td width="90%" height="90%" align="center" valign="top">

				<table><tr><td align='left'>
<!-- Wiki content -->

[% page.body %]

<!-- End wiki content -->

				</tr></td></table>

			</td>
			<td width="5%">&nbsp;</td>
		</tr>
		<tr>
			<td colspan="3" align="center" valign="middle">
				[% IF pages.previous -%]
					<a href='[% url(pages.previous) %]' class="nav">&lt;</a>
				[%- ELSE -%]
					<span class="nav">&lt;</span>
				[%- END %]

				<a href='[% url(pages.index) %]' class="nav">+</a>

				[% IF pages.next -%]
					<a href='[% url(pages.next) %]' class="nav">&gt;</a>
				[%- ELSE -%]
					<span class="nav">&gt;</span>
				[%- END %]
			</td>
		</tr>
	</table>

	</body>

</html>
